@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2006
  * Free Software Foundation, Inc.
  * Written by Steve Byrne and Paolo Bonzini.
  *
@@ -18,6 +18,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -25,7 +47,7 @@
  * 
  * You should have received a copy of the GNU General Public License along with
  * GNU Smalltalk; see the file COPYING.  If not, write to the Free Software
- * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
+ * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  
  *
  ***********************************************************************/
 
@@ -43,26 +65,28 @@
 /* The number of OOPs that are swept on each incremental GC step.  */
 #define INCREMENTAL_SWEEP_STEP	  100
 
+/* Define this flag to turn on debugging dumps for garbage collection */
+/* #define GC_DEBUG_OUTPUT */
+
 /* Define this flag to turn on debugging code for OOP table management */
 /* #define GC_DEBUGGING */
 
 /* Define this flag to disable incremental garbage collection */
 /* #define NO_INCREMENTAL_GC */
 
-/* Define this flag to disable blacking of grey pages (that is, the
-   entire oldspace is scanned to look for reachable newspace objects).
-   This is also necessary to run valgrind on GNU Smalltalk.  */
-/* #define NO_SIGSEGV_HANDLING */
-
 /* Define this flag to turn on debugging code for oldspace management */
-/* #define MMAN_DEBUGGING */
+/* #define MMAN_DEBUG_OUTPUT */
 
 #if !defined HAVE_SIGSEGV_RECOVERY || !(HAVE_SIGSEGV_RECOVERY-0)
 #define NO_SIGSEGV_HANDLING
 #endif
 
-#if defined(GC_DEBUGGING)
-#undef OPTIMIZE
+#if defined(GC_DEBUG_OUTPUT)
+#define GC_DEBUGGING
+#endif
+
+#if !defined(OPTIMIZE)
+#define GC_DEBUGGING
 #endif
 
 
@@ -74,7 +98,7 @@ OOP _gst_true_oop = NULL;
 OOP _gst_false_oop = NULL;
 
 /* This is true to show a message whenever a GC happens.  */
-mst_Boolean _gst_gc_message = true;
+int _gst_gc_message = true;
 
 /* This is != 0 in the middle of a GC.  */
 int _gst_gc_running = 0;
@@ -85,7 +109,7 @@ static heap oop_heap;
 /* This vector holds the storage for all the Character objects in the
    system.  Since all character objects are unique, we pre-allocate
    space for 256 of them, and treat them as special built-ins when
-   doing garbage collection.*/
+   doing garbage collection.  */
 static struct gst_character _gst_char_object_table[NUM_CHAR_OBJECTS];
 
 /* This is "nil" object in the system.  That is, the single instance
@@ -173,8 +197,11 @@ static void scan_grey_objects ();
    until no new object is found in it.  */
 static void scan_grey_pages ();
 
-/* Greys N pointers starting at BASE.  */
-static void add_to_grey_list (OOP *base, int n, OOP oop);
+/* Greys a page worth of pointers starting at BASE.  */
+static void add_to_grey_list (OOP *base, int n);
+
+/* Greys the object OOP.  */
+static void add_grey_object (OOP oop);
  
 /* Do the breadth-first scanning of copied objects.  */
 static void cheney_scan (void);
@@ -195,7 +222,7 @@ static void oldspace_nomemory (heap_data *h, size_t sz);
 
 /* Answer the number of fields to be scanned in the object starting
    at OBJ, with the given FLAGS on its OOP.  */
-static int scanned_fields_in (mst_Object obj, int flags) ATTRIBUTE_PURE;
+static int scanned_fields_in (gst_object obj, int flags) ATTRIBUTE_PURE;
 
 /* The mark phase of oldspace GC.  */
 static inline void mark_oops (void);
@@ -343,7 +370,7 @@ void _gst_update_object_memory_oop (OOP oop)
     numScavenges = _gst_mem.numScavenges;
     data = (gst_object_memory) OOP_TO_OBJ (oop);
     data->bytesPerOOP = FROM_INT (sizeof (PTR));
-    data->bytesPerOTE = FROM_INT (sizeof (struct OOP) +
+    data->bytesPerOTE = FROM_INT (sizeof (struct oop_s) +
 				  sizeof (gst_object_header));
 
     data->edenSize = FROM_INT (_gst_mem.eden.totalSize);
@@ -394,13 +421,13 @@ void _gst_update_object_memory_oop (OOP oop)
 }
 
 void
-_gst_init_oop_table (size_t size)
+_gst_init_oop_table (PTR address, size_t size)
 {
   int i;
 
   oop_heap = NULL;
   for (i = MAX_OOP_TABLE_SIZE; i && !oop_heap; i >>= 1)
-    oop_heap = _gst_heap_create (i * sizeof (struct OOP));
+    oop_heap = _gst_heap_create (address, i * sizeof (struct oop_s));
 
   if (!oop_heap)
     nomemory (true);
@@ -408,14 +435,14 @@ _gst_init_oop_table (size_t size)
   alloc_oop_table (size);
 
   _gst_nil_oop->flags = F_READONLY | F_OLD | F_REACHABLE;
-  _gst_nil_oop->object = (mst_Object) & _gst_nil_object;
+  _gst_nil_oop->object = (gst_object) & _gst_nil_object;
   _gst_nil_object.objSize =
     FROM_INT (ROUNDED_WORDS (sizeof (struct gst_undefined_object)));
 
   _gst_true_oop->flags = F_READONLY | F_OLD | F_REACHABLE;
-  _gst_true_oop->object = (mst_Object) & _gst_boolean_objects[0];
+  _gst_true_oop->object = (gst_object) & _gst_boolean_objects[0];
   _gst_false_oop->flags = F_READONLY | F_OLD | F_REACHABLE;
-  _gst_false_oop->object = (mst_Object) & _gst_boolean_objects[1];
+  _gst_false_oop->object = (gst_object) & _gst_boolean_objects[1];
   _gst_boolean_objects[0].objSize =
     FROM_INT (ROUNDED_WORDS (sizeof (struct gst_boolean)));
   _gst_boolean_objects[1].objSize =
@@ -429,7 +456,7 @@ _gst_init_oop_table (size_t size)
 	FROM_INT (ROUNDED_WORDS (sizeof (struct gst_character)));
       _gst_char_object_table[i].charVal = FROM_INT (i);
       _gst_mem.ot[i + CHAR_OBJECT_BASE].object =
-	(mst_Object) & _gst_char_object_table[i];
+	(gst_object) & _gst_char_object_table[i];
       _gst_mem.ot[i + CHAR_OBJECT_BASE].flags =
 	F_READONLY | F_OLD | F_REACHABLE;
     }
@@ -440,12 +467,11 @@ void
 alloc_oop_table (size_t size)
 {
   size_t bytes;
-  OOP oop;
 
   _gst_mem.ot_size = size;
-  bytes = (size - FIRST_OOP_INDEX) * sizeof (struct OOP);
+  bytes = (size - FIRST_OOP_INDEX) * sizeof (struct oop_s);
   _gst_mem.ot_base =
-    (struct OOP *) _gst_heap_sbrk (oop_heap, bytes);
+    (struct oop_s *) _gst_heap_sbrk (oop_heap, bytes);
   if (!_gst_mem.ot_base)
     nomemory (true);
 
@@ -455,16 +481,6 @@ alloc_oop_table (size_t size)
   _gst_false_oop = &_gst_mem.ot[FALSE_OOP_INDEX];
 
   _gst_mem.num_free_oops = size;
-
-  /* mark the OOPs as available */
-  PREFETCH_START (_gst_mem.ot, PREF_WRITE | PREF_NTA);
-  for (oop = _gst_mem.ot;
-       oop < &_gst_mem.ot[_gst_mem.ot_size]; oop++)
-    {
-      PREFETCH_LOOP (oop, PREF_WRITE | PREF_NTA);
-      oop->flags = F_FREE;
-    }
-
   _gst_mem.first_allocated_oop = _gst_mem.ot;
   _gst_mem.last_allocated_oop = _gst_mem.last_swept_oop = _gst_mem.ot - 1;
   _gst_mem.highest_swept_oop = _gst_mem.ot;
@@ -474,9 +490,8 @@ mst_Boolean
 _gst_realloc_oop_table (size_t newSize)
 {
   size_t bytes;
-  OOP oop;
 
-  bytes = (newSize - _gst_mem.ot_size) * sizeof (struct OOP);
+  bytes = (newSize - _gst_mem.ot_size) * sizeof (struct oop_s);
   if (bytes < 0)
     return (true);
 
@@ -488,15 +503,6 @@ _gst_realloc_oop_table (size_t newSize)
       return (false);
     }
 
-  /* mark the new OOPs as available */
-  PREFETCH_START (&_gst_mem.ot[_gst_mem.ot_size], PREF_WRITE | PREF_NTA);
-  for (oop = &_gst_mem.ot[_gst_mem.ot_size];
-       oop < &_gst_mem.ot[newSize]; oop++)
-    {
-      PREFETCH_LOOP (oop, PREF_WRITE | PREF_NTA);
-      oop->flags = F_FREE;
-    }
-
   _gst_mem.num_free_oops += newSize - _gst_mem.ot_size;
   _gst_mem.ot_size = newSize;
   return (true);
@@ -516,6 +522,55 @@ _gst_dump_oop_table()
           _gst_display_oop_short (oop);
       }
 }
+
+void
+_gst_dump_owners (OOP oop)
+{
+  OOP oop2, lastOOP;
+
+  for (oop2 = _gst_mem.ot_base, lastOOP = &_gst_mem.ot[_gst_mem.ot_size];
+       oop2 < lastOOP; oop2++)
+    if UNCOMMON (IS_OOP_VALID (oop2) && is_owner(oop2, oop))
+      _gst_display_oop (oop2);
+}
+
+void
+_gst_check_oop_table ()
+{
+  OOP oop, lastOOP;
+
+  for (oop = _gst_mem.ot_base, lastOOP = &_gst_mem.ot[_gst_mem.ot_size];
+       oop < lastOOP; oop++)
+    {
+      gst_object object;
+      OOP *scanPtr;
+      int n;
+
+      if (!IS_OOP_VALID_GC (oop))
+	continue;
+
+      object = OOP_TO_OBJ (oop);
+      scanPtr = &object->objClass;
+      if (oop->flags & F_CONTEXT)
+        {
+          gst_method_context ctx;
+          intptr_t methodSP;
+          ctx = (gst_method_context) object;
+          methodSP = TO_INT (ctx->spOffset);
+          n = ctx->contextStack + methodSP + 1 - object->data;
+        }
+      else
+        n = NUM_OOPS (object) + 1;
+
+      while (n--)
+        {
+	  OOP pointedOOP = *scanPtr++;
+	  if (IS_OOP (pointedOOP)
+	      && (!IS_OOP_ADDR (pointedOOP) || !IS_OOP_VALID_GC (pointedOOP)))
+            abort ();
+	}
+    }
+}
 
 
 void
@@ -613,7 +668,7 @@ void
 _gst_swap_objects (OOP oop1,
 		   OOP oop2)
 {
-  struct OOP tempOOP;
+  struct oop_s tempOOP;
   if (oop2->flags & F_WEAK)
     _gst_make_oop_non_weak (oop2);
 
@@ -661,23 +716,27 @@ _gst_swap_objects (OOP oop1,
 void
 _gst_make_oop_fixed (OOP oop)
 {
-  mst_Object newObj;
+  gst_object newObj;
   int size;
   if (oop->flags & F_FIXED)
     return;
 
-  size = SIZE_TO_BYTES (TO_INT(oop->object->objSize));
-  newObj = (mst_Object) _gst_mem_alloc (_gst_mem.fixed, size);
-  if (!newObj)
-    abort ();
+  if ((oop->flags & F_LOADED) == 0)
+    {
+      size = SIZE_TO_BYTES (TO_INT(oop->object->objSize));
+      newObj = (gst_object) _gst_mem_alloc (_gst_mem.fixed, size);
+      if (!newObj)
+        abort ();
 
-  memcpy (newObj, oop->object, size);
-  if (oop->flags & F_OLD)
-    _gst_mem_free (_gst_mem.old, oop->object);
-  else
-    _gst_mem.numOldOOPs++;
+      memcpy (newObj, oop->object, size);
+      if ((oop->flags & F_OLD) == 0)
+	_gst_mem.numOldOOPs++;
+      else
+        _gst_mem_free (_gst_mem.old, oop->object);
+
+      oop->object = newObj;
+    }
 
-  oop->object = newObj;
   oop->flags &= ~(F_SPACES | F_POOLED);
   oop->flags |= F_OLD | F_FIXED;
 }
@@ -685,19 +744,18 @@ _gst_make_oop_fixed (OOP oop)
 void
 _gst_tenure_oop (OOP oop)
 {
-  mst_Object newObj;
+  gst_object newObj;
   if (oop->flags & F_OLD)
     return;
 
   if (!(oop->flags & F_FIXED))
     {
       int size = SIZE_TO_BYTES (TO_INT(oop->object->objSize));
-      newObj = (mst_Object) _gst_mem_alloc (_gst_mem.old, size);
+      newObj = (gst_object) _gst_mem_alloc (_gst_mem.old, size);
       if (!newObj)
         abort ();
 
       memcpy (newObj, oop->object, size);
-
       _gst_mem.numOldOOPs++;
 
       oop->object = newObj;
@@ -708,12 +766,12 @@ _gst_tenure_oop (OOP oop)
 }
 
 
-mst_Object
+gst_object
 _gst_alloc_obj (size_t size,
 		OOP *p_oop)
 {
   OOP *newAllocPtr;
-  mst_Object p_instance;
+  gst_object p_instance;
 
   size = ROUNDED_BYTES (size);
 
@@ -730,7 +788,7 @@ _gst_alloc_obj (size_t size,
       newAllocPtr = _gst_mem.eden.allocPtr + size;
     }
 
-  p_instance = (mst_Object) _gst_mem.eden.allocPtr;
+  p_instance = (gst_object) _gst_mem.eden.allocPtr;
   _gst_mem.eden.allocPtr = newAllocPtr;
   *p_oop = alloc_oop (p_instance, _gst_mem.active_flag);
   p_instance->objSize = FROM_INT (BYTES_TO_SIZE (size));
@@ -738,28 +796,28 @@ _gst_alloc_obj (size_t size,
   return p_instance;
 }
 
-mst_Object
+gst_object
 _gst_alloc_old_obj (size_t size,
 		    OOP *p_oop)
 {
   OOP *newAllocPtr;
-  mst_Object p_instance;
+  gst_object p_instance;
 
   size = ROUNDED_BYTES (size);
   newAllocPtr = _gst_mem.eden.allocPtr + BYTES_TO_SIZE (size);
 
   /* If the object is big enough, we put it directly in oldspace.  */
-  p_instance = (mst_Object) _gst_mem_alloc (_gst_mem.old, size);
+  p_instance = (gst_object) _gst_mem_alloc (_gst_mem.old, size);
   if COMMON (p_instance)
     goto ok;
 
   _gst_global_gc (size);
-  p_instance = (mst_Object) _gst_mem_alloc (_gst_mem.old, size);
+  p_instance = (gst_object) _gst_mem_alloc (_gst_mem.old, size);
   if COMMON (p_instance)
     goto ok;
 
   _gst_compact (0);
-  p_instance = (mst_Object) _gst_mem_alloc (_gst_mem.old, size);
+  p_instance = (gst_object) _gst_mem_alloc (_gst_mem.old, size);
   if UNCOMMON (!p_instance)
     {
       /* !!! do something more reasonable in the future */
@@ -773,10 +831,10 @@ ok:
   return p_instance;
 }
 
-mst_Object _gst_alloc_words (size_t size)
+gst_object _gst_alloc_words (size_t size)
 {
   OOP *newAllocPtr;
-  mst_Object p_instance;
+  gst_object p_instance;
 
   /* We don't want to have allocPtr pointing to the wrong thing during
      GC, so we use a local var to hold its new value */
@@ -791,7 +849,7 @@ mst_Object _gst_alloc_words (size_t size)
   if UNCOMMON (size >= _gst_mem.big_object_threshold)
     abort ();
 
-  p_instance = (mst_Object) _gst_mem.eden.allocPtr;
+  p_instance = (gst_object) _gst_mem.eden.allocPtr;
   _gst_mem.eden.allocPtr = newAllocPtr;
   p_instance->objSize = FROM_INT (size);
   return p_instance;
@@ -808,11 +866,11 @@ reset_survivor_space (surv_space *space)
 void
 oldspace_after_allocating (heap_data *h, heap_block *blk, size_t sz)
 {
-#ifdef MMAN_DEBUGGING
+#ifdef MMAN_DEBUG_OUTPUT
   printf ("Allocating oldspace page at %p (%d)\n", blk, sz);
 #endif
 
-  add_to_grey_list ((OOP *) blk, sz / sizeof (PTR), NULL);
+  add_to_grey_list ((OOP *) blk, sz / sizeof (PTR));
   _gst_mem.rememberedTableEntries++;
 }
 
@@ -821,7 +879,7 @@ oldspace_before_freeing (heap_data *h, heap_block *blk, size_t sz)
 {
   grey_area_node *node, *last, **next;
 
-#ifdef MMAN_DEBUGGING
+#ifdef MMAN_DEBUG_OUTPUT
   printf ("Freeing oldspace page at %p (%d)\n", blk, sz);
 #endif
   
@@ -830,7 +888,7 @@ oldspace_before_freeing (heap_data *h, heap_block *blk, size_t sz)
     if (node->base >= (OOP *)blk
 	&& node->base + node->n <= (OOP *)( ((char *)blk) + sz))
       {
-#ifdef MMAN_DEBUGGING
+#ifdef MMAN_DEBUG_OUTPUT
 	printf ("  Remembered table entry removed %p..%p\n",
 		node->base, node->base+node->n);
 #endif
@@ -875,9 +933,9 @@ int oldspace_sigsegv_handler (void* fault_address, int serious)
   page = (char *) fault_address - ((intptr_t) fault_address & (getpagesize() - 1));
   errno = 0;
   if (_gst_mem_protect (page, getpagesize(), PROT_READ | PROT_WRITE) == -1 &&
-      (errno == ENOMEM || errno == EFAULT))
+      (errno == ENOMEM || errno == EFAULT || errno == EACCES || errno == EINVAL))
     {
-#if defined (MMAN_DEBUGGING)
+#if defined (MMAN_DEBUG_OUTPUT)
       printf ("Plain old segmentation violation -- address = %p\n", page);
 #endif
       reentering = 0;
@@ -888,12 +946,12 @@ int oldspace_sigsegv_handler (void* fault_address, int serious)
   (void) *(volatile char *) fault_address;
   reentering = 0;
 
-#if defined (MMAN_DEBUGGING)
+#if defined (MMAN_DEBUG_OUTPUT)
   printf ("Unprotected %p (SIGSEGV at %p)\n", page, fault_address);
 #endif
 
   _gst_mem.rememberedTableEntries++;
-  add_to_grey_list ((PTR) page, getpagesize() / sizeof (PTR), NULL);
+  add_to_grey_list ((PTR) page, getpagesize() / sizeof (PTR));
   return !reentered;
 }
 #endif
@@ -935,7 +993,7 @@ void
 _gst_compact (size_t new_heap_limit)
 {
   OOP oop;
-  grey_area_node *node, *next;
+  grey_area_node *node, **next, *last;
   heap_data *new_heap = init_old_space (
     new_heap_limit ? new_heap_limit : _gst_mem.old->heap_limit);
 
@@ -953,15 +1011,26 @@ _gst_compact (size_t new_heap_limit)
       update_stats (&stats.timeOfLastCompaction, NULL, NULL);
     }
 
-  /* Discard all the grey pages.  */
-  for (node = _gst_mem.grey_pages.head; node; node = next)
-    {
-      next = node->next;
-      xfree (node);
-    }
-  _gst_mem.grey_pages.head = _gst_mem.grey_pages.tail = NULL;
-  _gst_mem.rememberedTableEntries = 0;
+  /* Leave only pages from the loaded image in the grey table.  */
+  for (last = NULL, next = &_gst_mem.grey_pages.head; (node = *next); )
+    if (node->base >= (OOP *)_gst_mem.loaded_base
+        && node->base < _gst_mem.loaded_end)
+      {
+#ifdef MMAN_DEBUG_OUTPUT
+        printf ("  Remembered table entry left for loaded image: %p..%p\n",
+                node->base, node->base+node->n);
+#endif
+        last = node;
+        next = &(node->next);
+      }
+    else
+      {
+        _gst_mem.rememberedTableEntries--;
+        *next = node->next;
+        xfree (node);
+      }
 
+  _gst_mem.grey_pages.tail = last;
   _gst_fixup_object_pointers ();
 
   /* Now do the copying loop which will compact oldspace.  */
@@ -970,9 +1039,9 @@ _gst_compact (size_t new_heap_limit)
        oop < &_gst_mem.ot[_gst_mem.ot_size]; oop++)
     {
       PREFETCH_LOOP (oop, PREF_READ | PREF_NTA);
-      if ((oop->flags & (F_OLD | F_FIXED)) == F_OLD)
+      if ((oop->flags & (F_OLD | F_FIXED | F_LOADED)) == F_OLD)
         {
-          mst_Object new;
+          gst_object new;
           size_t size = SIZE_TO_BYTES (TO_INT (oop->object->objSize));
           new = _gst_mem_alloc (new_heap, size);
           memcpy (new, oop->object, size);
@@ -1012,7 +1081,7 @@ _gst_global_gc (int next_allocation)
     
   if (!_gst_gc_running++
       && _gst_gc_message
-      && _gst_verbosity > 0
+      && _gst_verbosity >= 2
       && !_gst_regression_testing)
     {
       /* print the first part of this message before we finish
@@ -1038,6 +1107,9 @@ _gst_global_gc (int next_allocation)
   _gst_mem.live_flags |= F_REACHABLE;
   check_weak_refs ();
   _gst_restore_object_pointers ();
+#if defined (GC_DEBUGGING)
+  _gst_check_oop_table ();
+#endif
   reset_incremental_gc (_gst_mem.ot);
 
   update_stats (&stats.timeOfLastGlobalGC,
@@ -1079,7 +1151,7 @@ _gst_global_gc (int next_allocation)
 
   if (!--_gst_gc_running
       && _gst_gc_message
-      && _gst_verbosity > 0
+      && _gst_verbosity >= 2
       && !_gst_regression_testing)
     {
       fprintf (stderr, "%s\"\n", s);
@@ -1112,7 +1184,7 @@ _gst_scavenge (void)
 
   if (!_gst_gc_running++
       && _gst_gc_message
-      && _gst_verbosity > 0
+      && _gst_verbosity > 2
       && !_gst_regression_testing)
     {
       /* print the first part of this message before we finish
@@ -1152,7 +1224,7 @@ _gst_scavenge (void)
 
   if (!--_gst_gc_running
       && _gst_gc_message
-      && _gst_verbosity > 0
+      && _gst_verbosity > 2
       && !_gst_regression_testing)
     {
       fprintf (stderr, "%d%% reclaimed, done\"\n", reclaimedPercent);
@@ -1182,7 +1254,7 @@ _gst_finish_incremental_gc ()
 {
   OOP oop, firstOOP;
 
-#if defined (GC_DEBUGGING)
+#if defined (GC_DEBUG_OUTPUT)
   printf ("Completing sweep (%p...%p), validity flags %x\n", _gst_mem.last_swept_oop,
           _gst_mem.highest_swept_oop, _gst_mem.live_flags);
 #endif
@@ -1304,7 +1376,7 @@ reset_incremental_gc (OOP firstOOP)
     _gst_realloc_oop_table (_gst_mem.ot_size
       * (100 + _gst_mem.space_grow_rate) / 100);
 
-#if defined (GC_DEBUGGING)
+#if defined (GC_DEBUG_OUTPUT)
   printf ("Found unallocated at OOP %p, last allocated OOP %p\n"
           "Highest OOP swept top to bottom %p, lowest swept bottom to top %p\n",
 	  _gst_mem.first_allocated_oop, _gst_mem.last_allocated_oop,
@@ -1323,12 +1395,10 @@ _gst_sweep_oop (OOP oop)
 
 #ifdef ENABLE_JIT_TRANSLATION
   if (oop->flags & F_XLAT)
-    /* Unreachable, always free the native code.  Note
-       it is *not* optional to free the code in this case -- and
-       I'm not talking about memory leaks: a different
-       method could use the same OOP as this one and the
-       old method one would be executed instead of the new
-       one! */
+    /* Unreachable, always free the native code.  It is *not* optional
+       to free the code in this case -- and I'm not talking about memory
+       leaks: a different method could use the same OOP as this one and
+       the old method would be executed instead of the new one! */
     _gst_release_native_code (oop);
 #endif
 
@@ -1341,23 +1411,25 @@ _gst_sweep_oop (OOP oop)
       _gst_mem.numOldOOPs--;
       stats.reclaimedOldSpaceBytesSinceLastGlobalGC +=
 	SIZE_TO_BYTES (TO_INT (OOP_TO_OBJ (oop)->objSize));
-      _gst_mem_free (_gst_mem.fixed, oop->object);
+      if ((oop->flags & F_LOADED) == 0)
+        _gst_mem_free (_gst_mem.fixed, oop->object);
     }
   else if UNCOMMON (oop->flags & F_OLD)
     {
       _gst_mem.numOldOOPs--;
       stats.reclaimedOldSpaceBytesSinceLastGlobalGC +=
 	SIZE_TO_BYTES (TO_INT (OOP_TO_OBJ (oop)->objSize));
-      _gst_mem_free (_gst_mem.old, oop->object);
+      if ((oop->flags & F_LOADED) == 0)
+        _gst_mem_free (_gst_mem.old, oop->object);
     }
 
-  oop->flags = F_FREE;
+  oop->flags = 0;
 }
 
 void
 mourn_objects (void)
 {
-  mst_Object array;
+  gst_object array;
   long size;
   gst_processor_scheduler processor;
 
@@ -1409,7 +1481,7 @@ queue_put (surv_space *q, OOP *src, int n)
       result = q->allocPtr;
       newAlloc = q->allocPtr + n;
 
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
       printf ("Top %p alloc %p tenure %p\n", q->topPtr, q->allocPtr, q->tenurePtr);
 #endif
 
@@ -1417,7 +1489,7 @@ queue_put (surv_space *q, OOP *src, int n)
         /* We tenure old objects as we copy more objects into
            the circular survivor space.  */
         {
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
 	  printf ("Tenure: current max %p, needed %p\n", q->tenurePtr, newAlloc);
 #endif
 	  tenure_one_object();
@@ -1426,7 +1498,7 @@ queue_put (surv_space *q, OOP *src, int n)
 
       if UNCOMMON (newAlloc > q->maxPtr)
         {
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
           printf ("Wrap: survivor space ends at %p, needed %p\n", q->maxPtr, newAlloc);
 #endif
           q->topPtr = q->allocPtr;
@@ -1458,14 +1530,14 @@ tenure_one_object ()
   OOP oop;
 
   oop = *_gst_mem.tenuring_queue.tenurePtr;
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
   printf ("      ");
   _gst_display_oop (oop);
 #endif
 
   if (_gst_mem.scan.current == oop)
     {
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
       printf ("Tenured OOP %p was being scanned\n", oop);
 #endif
 
@@ -1475,10 +1547,7 @@ tenure_one_object ()
 
   else if (_gst_mem.scan.queue_at == _gst_mem.tenuring_queue.tenurePtr)
     {
-      mst_Object obj;
-      int numFields;
-
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
       printf ("Tenured OOP %p had not been scanned yet\n", oop);
 #endif
 
@@ -1491,10 +1560,7 @@ tenure_one_object ()
         _gst_mem.scan.queue_at = _gst_mem.tenuring_queue.minPtr;
 
       _gst_tenure_oop (oop);
-      obj = OOP_TO_OBJ (oop);
-      numFields = scanned_fields_in (obj, oop->flags);
-      if (numFields)
-        add_to_grey_list (&(obj->objClass), numFields, oop);
+      add_grey_object (oop);
     }
   else
     _gst_tenure_oop (oop);
@@ -1517,27 +1583,50 @@ _gst_grey_oop_range (PTR from, size_t size)
 
 
 void
-add_to_grey_list (OOP *base, int n, OOP oop)
+add_grey_object (OOP oop)
 {
   grey_area_node *entry;
-  grey_area_list *p;
+  gst_object obj = OOP_TO_OBJ (oop);
+  int numFields = scanned_fields_in (obj, oop->flags);
+  OOP *base = &(obj->objClass);
 
-  if (oop)
-    p = &_gst_mem.grey_areas;
-  else
-    p = &_gst_mem.grey_pages;
+  if (!numFields)
+    return;
+
+  /* For ephemeron, skip the first field and the class.  */
+  if (oop->flags & F_EPHEMERON)
+    {
+      numFields -= &(obj->data[1]) - base;
+      base = &(obj->data[1]); 
+    }
 
   entry = (grey_area_node *) xmalloc (sizeof (grey_area_node));
   entry->base = base;
-  entry->n = n;
+  entry->n = numFields;
   entry->oop = oop;
   entry->next = NULL;
-  if (p->tail)
-    p->tail->next = entry;
+  if (_gst_mem.grey_areas.tail)
+    _gst_mem.grey_areas.tail->next = entry;
   else
-    p->head = entry;
+    _gst_mem.grey_areas.head = entry;
 
-  p->tail = entry;
+  _gst_mem.grey_areas.tail = entry;
+}
+
+void
+add_to_grey_list (OOP *base, int n)
+{
+  grey_area_node *entry = (grey_area_node *) xmalloc (sizeof (grey_area_node));
+  entry->base = base;
+  entry->n = n;
+  entry->oop = NULL;
+  entry->next = NULL;
+  if (_gst_mem.grey_pages.tail)
+    _gst_mem.grey_pages.tail->next = entry;
+  else
+    _gst_mem.grey_pages.head = entry;
+
+  _gst_mem.grey_pages.tail = entry;
 }
 
 void
@@ -1621,31 +1710,57 @@ copy_oops (void)
 }
 
 void
-scan_grey_pages ()
+_gst_print_grey_list (mst_Boolean check_pointers)
 {
-  grey_area_node *node, **next, *last;
+  grey_area_node *node;
   OOP *pOOP, oop;
   int i, n;
 
-#if defined (MMAN_DEBUGGING)
-  printf ("Pages on the grey list: ");
   for (n = 0, node = _gst_mem.grey_pages.head; node; node = node->next, n++)
     {
-      if (n & 3)
-        printf ("%12p ", node->base);
-      else
-        printf ("\n%12p ", node->base);
+      int new_pointers = 0;
+      if (check_pointers)
+        for (new_pointers = 0, pOOP = node->base, i = node->n; i--; pOOP++)
+          {
+            PREFETCH_LOOP (pOOP, PREF_READ | PREF_NTA);
+            oop = *pOOP;
+
+            /* Not all addresses are known to contain valid OOPs! */
+	    if (!IS_OOP_ADDR (oop))
+	      continue;
+
+            if (!IS_OOP_NEW (oop))
+	      continue;
+
+	    new_pointers++;
+	  }
+
+      printf ("%11p%c ", node->base, new_pointers == 0 ? ' ' : '*');
+      if ((n & 3) == 3)
+	putchar ('\n');
     }
 
   if (_gst_mem.grey_pages.tail)
     printf ("  (tail = %12p)", _gst_mem.grey_pages.tail->base);
 
   printf ("\n");
+}
+
+void
+scan_grey_pages ()
+{
+  grey_area_node *node, **next, *last;
+  OOP *pOOP, oop;
+  int i, n;
+
+#if defined (MMAN_DEBUG_OUTPUT)
+  printf ("Pages on the grey list:\n");
+  _gst_print_grey_list (true);
 #endif
 
   for (last = NULL, next = &_gst_mem.grey_pages.head; (node = *next); )
     {
-#if defined(GC_DEBUGGING) || defined(MMAN_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT) || defined(MMAN_DEBUG_OUTPUT)
       printf ("Scanning grey page %p...%p\n", node->base, node->base + node->n);
 #endif
 
@@ -1672,7 +1787,7 @@ scan_grey_pages ()
 	{
           /* The entry was temporary, or we found no new-space
              pointers in it.  Delete it and make the page read-only.  */
-#if defined (MMAN_DEBUGGING)
+#if defined (MMAN_DEBUG_OUTPUT)
 	  printf ("Protecting %p\n", node->base);
 #endif
 	  _gst_mem.rememberedTableEntries--;
@@ -1692,20 +1807,9 @@ scan_grey_pages ()
 
   _gst_mem.grey_pages.tail = last;
 
-#if defined (MMAN_DEBUGGING)
-  printf ("Pages left on the grey list: ");
-  for (n = 0, node = _gst_mem.grey_pages.head; node; node = node->next, n++)
-    {
-      if (n & 3)
-        printf ("%12p ", node->base);
-      else
-        printf ("\n%12p ", node->base);
-    }
-
-  if (_gst_mem.grey_pages.tail)
-    printf ("  (tail = %12p)", _gst_mem.grey_pages.tail->base);
-
-  printf ("\n");
+#if defined (MMAN_DEBUG_OUTPUT)
+  printf ("Pages left on the grey list:\n");
+  _gst_print_grey_list (false);
 #endif
 }
 
@@ -1714,7 +1818,7 @@ scan_grey_objects()
 {
   grey_area_node *node, *next;
   OOP oop;
-  mst_Object obj;
+  gst_object obj;
 
   for (next = _gst_mem.grey_areas.head; (node = next); )
     {
@@ -1725,7 +1829,7 @@ scan_grey_objects()
 	/* Objects might have moved, so update node->base.  */
 	node->base = (OOP *) &obj->data[1];
 
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
       printf ("Scanning grey range %p...%p (%p)\n", node->base, node->base + node->n, oop);
 #endif
 
@@ -1758,7 +1862,7 @@ scan_grey_objects()
 }
 
 int
-scanned_fields_in (mst_Object object,
+scanned_fields_in (gst_object object,
 		  int flags)
 {
   OOP objClass = object->objClass;
@@ -1793,7 +1897,7 @@ scanned_fields_in (mst_Object object,
 void
 cheney_scan (void)
 {
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
   printf ("Starting Cheney scan\n");
 #endif
 
@@ -1814,7 +1918,9 @@ cheney_scan (void)
 #if defined(GC_DEBUGGING)
       if (!IS_OOP_ADDR (oop))
 	abort();
+#endif
 
+#if defined(GC_DEBUG_OUTPUT)
       printf (">Scan ");
       _gst_display_oop (oop);
 #endif
@@ -1833,7 +1939,7 @@ cheney_scan (void)
 	MAYBE_COPY_OOP (_gst_mem.scan.at[i+1]);
     }
 
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
   printf ("Ending Cheney scan\n");
 #endif
 }
@@ -1852,18 +1958,18 @@ _gst_copy_an_oop (OOP oop)
   int i, n;
   do
     {
-      mst_Object obj;
+      gst_object obj;
       OOP *pData;
 
       obj = OOP_TO_OBJ (oop);
       pData = (OOP *) obj;
 
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
       printf (">Copy ");
       _gst_display_oop (oop);
 #endif
 
-#if !defined (OPTIMIZE)
+#if defined (GC_DEBUGGING)
       if UNCOMMON (!IS_INT (obj->objSize))
 	{
 	  printf ("Size not an integer in OOP %p (%p)\n", oop, obj);
@@ -1875,20 +1981,13 @@ _gst_copy_an_oop (OOP oop)
 	  printf ("Invalid size for OOP %p (%p)\n", oop, obj);
 	  abort ();
 	}
-#endif
 
-      if UNCOMMON (oop->flags & F_FREE)
+      if UNCOMMON (oop->flags == 0)
 	{
 	  printf ("Free OOP %p was referenced\n", oop);
-	  /* If we're lucky, we can print some information on it...
-	     if we're not, we segfault: not that bad given that
-	     we'll abort pretty soon! */
-	  oop->flags &= ~F_FREE;
-	  _gst_display_oop (oop);
 	  abort ();
 	}
 
-#if !defined (OPTIMIZE)
       if UNCOMMON ((oop->flags & F_OLD) ||
 	  IS_SURVIVOR_ADDR(obj, _gst_mem.active_half == &_gst_mem.surv[1]))
         {
@@ -1899,7 +1998,7 @@ _gst_copy_an_oop (OOP oop)
 #endif
 
       queue_put (&_gst_mem.tenuring_queue, &oop, 1);
-      obj = oop->object = (mst_Object)
+      obj = oop->object = (gst_object)
 	queue_put (_gst_mem.active_half, pData, TO_INT (obj->objSize));
 
       oop->flags &= ~(F_SPACES | F_POOLED);
@@ -1912,9 +2011,8 @@ _gst_copy_an_oop (OOP oop)
       n = scanned_fields_in (obj, oop->flags) - 1;
       if (oop->flags & F_EPHEMERON)
 	{
-	  /* For ephemerons, grey the instance variables after
-	     the first.  */
-          add_to_grey_list (&obj->data[1], n - 1, oop);
+	  /* For ephemerons, do the work later.  */
+          add_grey_object (oop);
 	  return;
 	}
 
@@ -1944,7 +2042,7 @@ mark_oops (void)
 void
 mark_ephemeron_oops (void)
 {
-  OOP *pOOP, *base;
+  OOP *pOOP, *pDeadOOP, *base;
   int i, size;
 
   /* Make a local copy of the buffer */
@@ -1959,7 +2057,7 @@ mark_ephemeron_oops (void)
   for (pOOP = base, i = size; i--; pOOP++)
     {
       OOP oop = *pOOP;
-      mst_Object obj = OOP_TO_OBJ(oop);
+      gst_object obj = OOP_TO_OBJ(oop);
       OOP key = obj->data[0];
 
       if (key->flags & F_REACHABLE)
@@ -1968,10 +2066,10 @@ mark_ephemeron_oops (void)
       key->flags |= F_REACHABLE;
     }
 
-  for (pOOP = base, i = size; i--; pOOP++)
+  for (pOOP = pDeadOOP = base, i = size; i--; )
     {
-      OOP oop = *pOOP;
-      mst_Object obj = OOP_TO_OBJ(oop);
+      OOP oop = *pOOP++;
+      gst_object obj = OOP_TO_OBJ(oop);
       OOP key = obj->data[0];
       int num = NUM_OOPS(obj);
       int j;
@@ -1983,10 +2081,10 @@ mark_ephemeron_oops (void)
       for (j = 1; j < num; j++)
         MAYBE_MARK_OOP (obj->data[j]);
 
-      /* Remember that above we cleared F_EPHEMERON is the key
+      /* Remember that above we cleared F_EPHEMERON if the key
          is alive.  */
       if (!IS_OOP_MARKED (key) && (oop->flags & F_EPHEMERON))
-        _gst_add_buf_pointer (oop);
+        *pDeadOOP++ = oop;
 
       /* Ok, now mark the key.  */
       MAYBE_MARK_OOP (key);
@@ -1994,6 +2092,12 @@ mark_ephemeron_oops (void)
       /* Restore the flag in case it was cleared.  */
       oop->flags |= F_EPHEMERON;
     }
+
+  /* If more ephemerons were reachable from the object, go on...  */
+  if (_gst_buffer_size ())
+    mark_ephemeron_oops ();
+
+  _gst_add_buf_data (base, (char *) pDeadOOP - (char *) base);
 }
 
 #define TAIL_MARK_OOP(newOOP) BEGIN_MACRO { \
@@ -2019,8 +2123,8 @@ _gst_mark_an_oop_internal (OOP oop,
     {
       if (!oop)
 	{			/* in the loop! */
-#if !defined (OPTIMIZE)
-	  mst_Object obj = (mst_Object) (curOOP - 1);	/* for debugging */
+#if defined (GC_DEBUGGING)
+	  gst_object obj = (gst_object) (curOOP - 1);	/* for debugging */
 #endif
 	iterationLoop:
 	  /* in a loop, do next iteration */
@@ -2029,7 +2133,7 @@ _gst_mark_an_oop_internal (OOP oop,
 	  curOOP++;
 	  if (IS_OOP (oop))
 	    {
-#if !defined (OPTIMIZE)
+#if defined (GC_DEBUGGING)
 	      if UNCOMMON (!IS_OOP_ADDR (oop))
 		{
 		  printf
@@ -2064,10 +2168,10 @@ _gst_mark_an_oop_internal (OOP oop,
       else
 	{			/* just starting with this oop */
 	  OOP objClass;
-	  mst_Object object;
+	  gst_object object;
 	  uintptr_t size;
 
-#if !defined (OPTIMIZE)
+#if defined (GC_DEBUGGING)
 	  if UNCOMMON (IS_OOP_FREE (oop))
 	    {
 	      printf ("Error! Free OOP %p is being marked!\n", oop);
@@ -2076,7 +2180,7 @@ _gst_mark_an_oop_internal (OOP oop,
 	    }
 #endif
 
-#if defined(GC_DEBUGGING)
+#if defined(GC_DEBUG_OUTPUT)
 	  printf (">Mark ");
 	  _gst_display_oop (oop);
 #endif
